@@ -6,25 +6,41 @@ use base qw/Class::Accessor/;
 use strict;
 use warnings;
 
-use Scalar::Util ();
+use Scalar::Util qw/blessed refaddr reftype/;
 use overload ();
 use Symbol ();
 
-our $VERSION = "0.04";
+our $VERSION = "0.07";
 
 sub visit {
 	my ( $self, $data ) = @_;
 
-	local $self->{_seen} = ($self->{_seen} || {});
-	return $data if ref $data and $self->{_seen}{ overload::StrVal( $data ) }++;
+	my $seen_hash = local $self->{_seen} = ($self->{_seen} || {}); # delete it after we're done with the whole visit
+	if ( ref $data ) { # only references need recursion checks
+		if ( exists $seen_hash->{ refaddr( $data ) } ) { # if it's been seen
+			return $seen_hash->{ refaddr( $data ) }; # return whatever it was mapped to
+		} else {
+			my $seen = \( $seen_hash->{ refaddr( $data ) } );
+			$$seen = $data;
 
-	if ( Scalar::Util::blessed( $data ) ) {
-		return $self->visit_object( $data );
-	} elsif ( my $reftype = ref $data ) {
-		if ( $reftype eq "HASH" or $reftype eq "ARRAY" or $reftype eq "GLOB" or $reftype eq "SCALAR") {
-			my $method = lc "visit_$reftype";
-			return $self->$method( $data );
+			if ( defined wantarray ) {
+				return $$seen = $self->visit_no_rec_check( $data );
+			} else {
+				return $self->visit_no_rec_check( $data );
+			}
 		}
+	} else {
+		return $self->visit_no_rec_check( $data );
+	}
+}
+
+sub visit_no_rec_check {
+	my ( $self, $data ) = @_;
+
+	if ( blessed( $data ) ) {
+		return $self->visit_object( $data );
+	} elsif ( ref $data ) {
+		return $self->visit_ref( $data );
 	}
 	
 	return $self->visit_value( $data );
@@ -36,6 +52,21 @@ sub visit_object {
 	return $self->visit_value( $object );
 }
 
+sub visit_ref {
+	my ( $self, $data ) = @_;
+
+	 my $reftype = reftype $data;
+
+	my $method = lc "visit_$reftype";
+
+	if ( $self->can($method) ) {
+		return $self->$method( $data );
+	} else {
+		return $self->visit_value($data);
+	}
+
+}
+
 sub visit_value {
 	my ( $self, $value ) = @_;
 
@@ -174,6 +205,10 @@ Then override the callback methods in any way you like. To retain visitor
 behavior, make sure to retain the functionality of C<visit_array> and
 C<visit_hash>.
 
+=head1 TODO
+
+Add support for "natural" visiting of trees.
+
 =head1 SEE ALSO
 
 L<Tree::Simple::VisitorFactory>, L<Data::Traverse>
